#include        "AvrX.inc"
/*
        Copyright 1999-2000, Larry Barello
        larry@barello.net

*/
        .section .text.avrx.SINGLESTEPNEXT

;+
;-----------------------------------------------------------
; uint8_t AvrXSingleStepNext(pProcessID)
;
; Marks a suspended task for single step support
; Jams it on the front of the run queue.  Returns error
; if task is not suspended.
;
; PASSED:       R25:R24 = Pid to single step
; RETURN:       R24 = 1 if task not suspended (error)
; USES:         FLAGS (except T)
;-
        .global AvrXSingleStepNext
        .func AvrXSingleStepNext
AvrXSingleStepNext:
        AVRX_Prolog

        mov     Zh, R25
        mov     Zl, R24
        ldd     Xl, Z+PidState
        sbrs    Xl, SuspendedBit
        rjmp    AssnXErr
        
        sbr     Xl, BV(SingleStep) ; Mark for single stepping
        std     Z+PidState, Xl
        
        lds     Xh, RunQueue   ; Put task in front of run queue
        lds     Xl, RunQueue+1
        sts     RunQueue, Zh
        sts     RunQueue+1, Zl
        std     Z+NextH, Xh
        std     Z+NextL, Xl
        rjmp    AssnXOk
        .endfunc

/*+
; -----------------------------------------------------------
; uint8_t AvrXStepNext(pProcessID)
;
; Unsuspends a task, adds it to the run queue
; then resuspends the task.  
;
; PASSED:       R25:R24 = Pid to single step
; RETURN:       R24 = 1 if task not suspended (error)
; USES:         
;
-*/
        .global AvrXStepNext
        .func AvrXStepNext
AvrXStepNext:
        AVRX_Prolog

        mov     Zh, R25
        mov     Zl, R24
        ldd     Xl, Z+PidState
        sbrs    Xl, SuspendedBit
        rjmp    AssnX
        
        cbr     Xl, BV(SuspendBit) | BV(SuspendedBit)
        std     Z+PidState, Xl
        rcall   _QueuePid               ; Add to run queue
; <BUG BUG - QueuePid trashes Z?, yeah, reloads with R25:R24
        ldd     Xl, Z+PidState
        sbr     Xl, BV(SuspendBit)
        std     Z+PidState, Xl
AssnXOk:
        ldi     R24, lo8(0)     ; Ok Return
        rjmp    AssnX
AssnXErr:
        ldi     R24, lo8(1)     ; Err Return
AssnX:
        std     Y+_R24, R24
        rjmp    _Epilog
        
        .endfunc


