#include        "AvrX.inc"
/*
        Copyright 1999-2000, Larry Barello
        larry@barello.net

*/
;
; AvrX Time Queue Manager
;

        .comm  _TimerQueue, 2       ; Head of the timer queue
        .comm  _TimQLevel, 1       ; Recurse level of timer interrupts

        .section .text.avrx.TIMEQUEUEMANAGER

/*+
; -----------------------------------------------
; void AvrXDelay(pTcb, unsigned)
;
; Passed:       R25:R24 = TCB
;               R23:R22 = Count
; Returns:      
; Uses:         
; Stack: 
; Notes:        See AvrXStartTimer, AvrXWaitTimer
-*/
        .global AvrXDelay
        .func   AvrXDelay
AvrXDelay:
        rcall   AvrXStartTimer
        rjmp    AvrXWaitTimer
        .endfunc
/*+
; -----------------------------------------------
; AvrXStartTimer
;
; Passed:       R25:R24 = TCB
;               R23:R22 = Count
; Returns:
; Uses:
; Stack:        
; Notes:        Should check and halt if TCB already queued
;               Resets TCB Semaphore  If count Zero, just flag
;               semaphore and return
-*/
        .global AvrXStartTimer
        .func   AvrXStartTimer
AvrXStartTimer:
        subi    R22, lo8(-0)
        sbci    R23, hi8(-0)
        brne    CountNotZero
        rjmp    AvrXSetObjectSemaphore
        .endfunc
        .global CountNotZero
        .func   CountNotZero
CountNotZero:        
        AVRX_Prolog

        rcall   AvrXResetObjectSemaphore

        ldi     Zl, lo8(_TimerQueue)
        ldi     Zh, hi8(_TimerQueue)
        
        BeginCritical
        lds     R0, _TimQLevel
        dec     R0
        sts     _TimQLevel, R0
        EndCritical

ast00:
        mov     Yl, Zl          ; Y -> Previous
        mov     Yh, Zh
        ldd     Zl, Y+NextL     ; Z -> Current
        ldd     Zh, Y+NextH
        adiw    Zl, 0
        breq    ast01           ; End of queue, Wrap up.

        ldd     Xl, Z+TcbCount+1; X = Current count
        ldd     Xh, Z+TcbCount  ; R23:22 = our count

        sub     R22, Xl
        sbc     R23, Xh         ; Subtract it from us
        brsh    ast00           ; Its less than us, continue walking the list

        add     R22, Xl         ; Restore us
        adc     R23, Xh
        sub     Xl, R22         ; Subtract us from it
        sbc     Xh, R23

        std     Z+TcbCount+1, Xl
        std     Z+TcbCount, Xh  ; Put it back out and insert us in front.
;
; Wrap up: Z->Current, R25:24->Tcb, Y->Prev, R23:22 = Count
; Insert Tcb into chain
;
ast01:
        std     Y+NextH, R25
        std     Y+NextL, R24    ; Prev.next = NewTCB
        mov     Yh, R25
        mov     Yl, R24
        std     Y+NextH, Zh     ; NewTCB.next = Current
        std     Y+NextL, Zl
        std     Y+TcbCount, R23 ; NewTCB.Count = count
        std     Y+TcbCount+1, R22

        rcall   _TimerHandler   ; process any nested timer interrupts

        rjmp    _Epilog
        .endfunc

/*+
; -----------------------------------------------
; AvrXTimerHandler
;
; THIS IS INTERRUPT CODE and must be called within an
; _Prolog/_Epilog section.
;
; The register variable _TimQLevel is used to count re-
; entry into the code.  It is also a semaphore that prevents
; timer interrupts from hosing the AvrXStartTimer code, above.
; Thus, the entire timer handler chain can be run with interrupts
; enabled.
;
; Since this can be called from C code (avrx 2.5) gotta preserve Y
; and R1-R16.  System calls within can and do trash the trashable
; registers, hence all the push/pop's
-*/
        .global AvrXTimerHandler
        .func   AvrXTimerHandler
AvrXTimerHandler:
        BeginCritical
        lds     Xl, _TimQLevel
        subi    Xl, 1           ; Can't use "dec" because doesn't affect 
        sts     _TimQLevel, Xl  ; carry flag.
        EndCritical
        brcs    ati00
        ret                     ; Do not branch on rollover (0->FF)
ati00:
        push    Yl              ; Gotta save these since we do not know
        push    Yh              ; who called us (C code)
        lds     Yh, _TimerQueue
        lds     Yl, _TimerQueue+1
        adiw    Yl, 0
        breq    ati02           ; Empty queue, return
;
; Y points to the first element in the queue.
;
        ldd     Zh, Y+TcbCount
        ldd     Zl, Y+TcbCount+1        
        sbiw    Zl, 1           ; Y->Count--
        std     Y+TcbCount, Zh
        std     Y+TcbCount+1, Zl
ati01:
        or      Zl, Zh          ; While (Y->Count == 0)
        brne    ati02           ; {

        ldd     Xl, Y+NextL
        ldd     Xh, Y+NextH
        sts     _TimerQueue, Xh ; Point queue to (new) first element
        sts     _TimerQueue+1, Xl
        std     Y+NextH, Zh     ;   Zero out link
        std     Y+NextL, Zl
        
        push    Xl
        push    Xh
        push    Zl
        push    Zh
        ldd     R25, Y+TcbSemaphore
        ldd     R24, Y+TcbSemaphore+1
        subi    R24, lo8(TIMERMESSAGE_EV)
        sbci    R25, hi8(TIMERMESSAGE_EV)
        brne    1f
        ldd     R24, Y+TcbQueue+1
        ldd     R25, Y+TcbQueue
        mov     R23, Yh
        mov     R22, Yl
        rcall   AvrXIntSendMessage
        rjmp    2f
1:
        mov     R24, Yl
        mov     R25, Yh
        rcall   AvrXIntSetObjectSemaphore
2:
        pop     Zh
        pop     Zl
        pop     Xh
        pop     Xl

        adiw    Xl, 0           ;   If Next Tcb == 0 (list empty)
        breq    ati02           ;       break;
                                ;   else
        mov     Yl, Xl
        mov     Yh, Xh
        ldd     Zh, Y+TcbCount  ;       load up count and loop.
        ldd     Zl, Y+TcbCount+1
        rjmp    ati01           ; }
ati02:
        pop     Yh
        pop     Yl
_TimerHandler:
        BeginCritical
        lds     R0, _TimQLevel
        inc     R0
        sts     _TimQLevel, R0
        EndCritical
        brne    ati00
        ret
        
        .endfunc

