/*
File:   Serialio.asm

Copyright 1998 - 2001 Larry Barello

Revision:

1/31/01 - Changed InitSerialIO parameter register to conform to new
        GCC calling conventions (R25 -> R24)

*/
#include        "avrx.inc"
#include        "serialio.inc"
;
; Needed for USART interrupt
;
;
#define UCR_INIT  lo8((1<<TXEN) | (1<<RXEN) | (1<<RXCIE))
// UBRR_INIT is an external value defined by user application.
//#define UBRR_INIT lo8((CPUCLK)/(16 * BAUDRATE) - 1)

#define CharPushed 7

        .lcomm  SioFlags,1
        .lcomm  _RxByte, 1
        .lcomm  RxC_Flag, MtxSz   
        .lcomm  TxIntFlag, MtxSz

        .section        .text.avrx.serialio
;+
;------------------------------------------
; PushChar
;
; Set a flag so that the next call to GetChar just returns.
; in effect pushing the last returned charactor to be retrieved
; by the next routine
;-
        .global PushChar
        .func   PushChar
PushChar:
        push    Xl
        lds     Xl, SioFlags
        sbr     Xl, 1<<CharPushed
        sts     SioFlags, Xl
        pop     Xl
        ret
        .endfunc
;+
;------------------------------------------
; GetChar
;
; Need to make this draw from a buffer so that multiple
; chars can be pushed... and so different input streams
; can be attached to the buffer.
;-
        .global GetChar
;        .global ?C_GETCHAR
        .func   GetChar
;?C_GETCHAR:
GetChar:
        push    Xl
        push    Xh
        push    Zl
        push    Zh
        push    R24
        push    R25
        lds     Xl, SioFlags
        sbrs    Xl, CharPushed
        rjmp    gc00
        cbr     Xl, 1<<CharPushed
        rjmp    gc01
        sts     SioFlags, Xl
        rjmp    gc01
gc00:
        ldi     R25, hi8(RxC_Flag)
        ldi     R24, lo8(RxC_Flag)
        rcall   AvrXWaitSemaphore       ; Wait for a charactor, or flag from System
gc01:
        lds     RxByte, _RxByte
        pop     R25
        pop     R24
        pop     Zh
        pop     Zl
        pop     Xh
        pop     Xl
        ret
        .endfunc
;+
;------------------------------------------
; _GetChar
;
; Special version that echos
;
; RETURNS:      RxByte (Register)
;-
        .global _GetChar
        .func   _GetChar
_GetChar:
        rcall   GetChar
        mov     TxByte, RxByte          ; Probably a no-op
        rjmp    PutChar
        .endfunc
;+
;------------------------------------------
;
; PutCR
;
; Put a CR to the output stream
;-
        .global PutCR
        .func   PutCR
PutCR:
        ldi     TxByte, 0x0D
        rjmp    PutChar
        .endfunc
;+
;------------------------------------------
;
; PutSpace
;
; Put a space to the output stream
;-
        .global PutSpace
        .func   PutSpace
PutSpace:
        ldi     TxByte, ' '
        rjmp    PutChar
        .endfunc
;+
;------------------------------------------
;
; PutString
;
; Prints out a string in code memory
;-
        .global _PrintString
        .func   _PrintString
_PrintString:
        lpm
        tst     R0
        breq    _ps00
        mov     TxByte, R0
        rcall   PutChar
        adiw    Zl, 1
        rjmp    _PrintString
_ps00:
        ret
        .endfunc
;+
;-------------------------------------------
; InitSerialIO
;
; PASSED:       R24 = UBBR
;
; History: parameter used to be passed in R25 l.b. 1/31/01
;
        .global InitSerialIO
        .func   InitSerialIO
InitSerialIO:
        out     UBRR, R24
        ldi     R31, UCR_INIT
        out     UCR, R31
        sbi     PORTD, PD0      ; Enable pull up on Rx
        ret
        .endfunc
;
;-----------------------------------------------
; UDRE_Interrupt
;
; Disables the interrupt (UDR is empty) signal waiting
; process to stuff the UDR
;
        .global _uart_data_
        .func   _uart_data_
_uart_data_:
        cbi     UCR, UDRIE      ; Disable the interrupt
        
        rcall   IntProlog
        
        ldi     R24, lo8(TxIntFlag)
        ldi     R25, hi8(TxIntFlag)
        rcall   AvrXIntSetSemaphore     ; Signal we are ready for more.
        
        rjmp    _Epilog
        .endfunc
        
;
;-----------------------------------------------
; RxC_Interrupt
;
; Reads the USART and flags UserTask, Char In semaphore
;
        .global _uart_recv_
        .func   _uart_recv_
_uart_recv_:
        cbi     UCR, RXCIE      ; disable interrupt

        rcall   IntProlog      ; Switch to kernel stack
        
        in      Xl, UDR
        sts     _RxByte, Xl

        ldi     R25, hi8(RxC_Flag)
        ldi     R24, lo8(RxC_Flag)
        rcall   AvrXIntSetSemaphore

        sbi     UCR, RXCIE      ; Re-enable interrupts
        rjmp    _Epilog
        .endfunc
;+
;------------------------------------------
; PutChar
;
; Stuffs "TxByte" into usart.  The usart is double buffered.
; when the data register is full we wait on the semaphore
; until the Usart interrupts indicating the register is ready.
;-
        .global PutChar
;        .global ?C_PUTCHAR
        .func   PutChar
;?C_PUTCHAR:
PutChar:
        push    Xl
        push    Xh
        push    Zl
        push    Zh
        push    R24
        push    R25
        sbic    USR, UDRE
        rjmp    pc00
        
        sbi     UCR, UDRIE      ; Enable interrupts and
        ldi     R24, lo8(TxIntFlag)
        ldi     R25, hi8(TxIntFlag)
        rcall   AvrXWaitSemaphore; Wait until data register is ready
pc00:
        out     UDR, TxByte

        pop     R25
        pop     R24
        pop     Zh
        pop     Zl
        pop     Xh
        pop     Xl
        ret
        .endfunc
        
        .section .text.avrx.siohelp
;-----------------------------------------
; _GetHex
;
; PASSED:       R24 = Maximum number of digits
; RETURNS:      Y = val, Tbit, R25 = Number of digits entered
; USES:         R24, R25
; STACK:        4
;
        .global _GetHex
        .func   _GetHex
_GetHex:
        clr     Yl
        clr     Yh
        clr     R25
gh_00:
        rcall   _GetChar
        rcall   IsHex
        brtc    gh_xx
        cpi     RxByte, '9'+1
        brge    gh_01
        subi    RxByte, '0'
        rjmp    gh_02
gh_01:
        cbr     RxByte, 0x20
        subi    RxByte, 'A'-10
gh_02:
        lsl     Yl
        rol     Yh
        lsl     Yl
        rol     Yh
        lsl     Yl
        rol     Yh
        lsl     Yl
        rol     Yh
        add     Yl, RxByte
        inc     R25
        dec     R24
        brne    gh_00
gh_xx:
        ret
        .endfunc


;-----------------------------------------
; GetHexWord
;
        .global GetHexWord
        .func   GetHexWord
GetHexWord:
        ldi     R24, 4
        rjmp    _GetHex
        .endfunc
;-----------------------------------------
; GetHexByte
;
        .global GetHexByte
        .func   GetHexByte
GetHexByte:
        ldi     R24, 2
        rjmp    _GetHex
        .endfunc
;-----------------------------------------
; GetHexByte
;
        .global GetHexChar
        .func   GetHexChar
GetHexChar:
        ldi     R24, 1
        rjmp    _GetHex
        .endfunc
;-----------------------------------------
; IsHex
;
; Tests RxByte for ascii hex char
; Returns: T bit
;
        .global IsHex
        .func   IsHex
IsHex:
        clt
        cpi     RxByte, '0'
        brlt    ihno
        cpi     RxByte, '9'+1
        brlt    ihyes
        cpi     RxByte, 'A'
        brlt    ihno
        cpi     RxByte, 'F'+1
        brlt    ihyes
        cpi     RxByte, 'a'
        brlt    ihno
        cpi     RxByte, 'f'+1
        brge    ihno
ihyes:  set
ihno:   ret
        .endfunc
;-----------------------------------------
; IsWhite
;
; Tests RxByte for space or CR
; Returns: T bit
;
        .global IsWhite
        .func   IsWhite
IsWhite:
        clt
        cpi     RxByte, ' '
        breq    ihyes
        cpi     RxByte, 0x0D
        breq    ihyes
        ret
        .endfunc
;-----------------------------------------
; _PutHex
; 
; Passed:       Y = Value, R24 = # digits
; RETURN:       Nothing
; USES:         R25, TxByte
; STACK:        2
;
        .global _PutHex
        .func   _PutHex
_PutHex:
        ldi     R25, 4          ; Maximum number of digits.
ph00:
        lsl     Yl
        rol     Yh
        rol     TxByte
        lsl     Yl
        rol     Yh
        rol     TxByte
        lsl     Yl
        rol     Yh
        rol     TxByte
        lsl     Yl
        rol     Yh
        rol     TxByte

        cp      R24, R25
        brlt    ph02

        cbr     TxByte, 0xF0
        subi    TxByte, lo8(-'0')       ; Addi 0
        cpi     TxByte, lo8('9' + 1)    ; if > 9
        brlt    ph01                    ; subi 9 and add A
        subi    TxByte, lo8('9' + 1 - 'A')
ph01:   
        rcall   PutChar
ph02:   
        dec     R25
        brne    ph00

        ret
        .endfunc
;-----------------------------------------
; Put out a hex word
; 
; Passed: Y = word
;
        .global PutHexWord
        .func   PutHexWord
PutHexWord:
        ldi     R24, 4
        rjmp    _PutHex
        .endfunc
;-----------------------------------------
; Put out a two ascii hex byte
; Yl = byte
;
        .global PutHexByte
        .func   PutHexByte
PutHexByte:
        ldi     R24, 2
        rjmp    _PutHex
        .endfunc

